/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: MicrosoftWindowsCertificateHandler.java,v 1.8 2006/08/13 21:21:24 cara Exp $ */

package org.openoces.opensign.certificate.plugin.capi;

/**
 * This class represents the interface to a certificate accessed through the Microsoft cryptoapi
 *
 * @author Carsten Raskgaard   <carsten@raskgaard.dk>
 * @author Preben Valeur  <prv@tdc.dk>
 */

import org.openoces.opensign.certificate.AbstractCertificateHandler;
import org.openoces.opensign.certificate.StringPrincipal;
import org.openoces.opensign.certificate.x509.KeyUsage;
import org.openoces.opensign.client.applet.bootstrap.interfaces.RestrictedAction;
import org.openoces.opensign.client.applet.bootstrap.FileLog;
import org.openoces.opensign.exceptions.UserCancel;
import org.openoces.opensign.utils.HexDump;
import org.openoces.opensign.wrappers.microsoftcryptoapi.MicrosoftCryptoApi;

import java.io.IOException;
import java.math.BigInteger;
import java.security.Principal;
import java.util.Date;
import java.util.Properties;

public class MicrosoftWindowsCertificateHandler implements AbstractCertificateHandler {

    byte[] binaryCertificate;
    MicrosoftCryptoApi capi;
    Principal subjectPrincipal;
    Principal issuerPrincipal;
    KeyUsage keyUsage;
    private RestrictedAction ra;

    public MicrosoftWindowsCertificateHandler(byte[] binaryCertificate, RestrictedAction ra) {
        this.binaryCertificate = binaryCertificate;
        this.ra = ra;
        capi = new MicrosoftCryptoApi();
        subjectPrincipal = null;
        keyUsage = null;
    }

    public boolean isInfoAvailable() {
        return true;
    }

    public String getUserfriendlyName() {
        // remove anything but CN= to shorten name as much as possible. Detailed info can be viewed by selecting details
        int s0 = getSubjectDN().getName().toLowerCase().indexOf("cn=") + 3;
        int s1 = getSubjectDN().getName().toLowerCase().indexOf(",", s0);
        int s2 = getSubjectDN().getName().toLowerCase().indexOf("+", s0);
        if (s1 == -1)
            s1 = getSubjectDN().getName().length();
        if (s2 != -1)
            s1 = s2;
        return getSubjectDN().getName().substring(s0, s1).trim();
    }

    public Principal getSubjectDN() {
        if (subjectPrincipal == null) {
            subjectPrincipal = new StringPrincipal(capi.getSubjectDnString(binaryCertificate));
        }
        return subjectPrincipal;
    }

    public Principal getIssuerDN() {
        if (issuerPrincipal == null) {
            issuerPrincipal = new StringPrincipal(capi.getIssuerDnString(binaryCertificate));
        }
        return issuerPrincipal;
    }

    public BigInteger getSerialNumber() {
        return capi.getSerialNumberBigInteger(binaryCertificate);
    }

    public Date getNotBefore() {
        return capi.getNotBeforeDate(binaryCertificate);
    }

    public Date getNotAfter() {
        return capi.getNotAfterDate(binaryCertificate);
    }

    public int getVersion() {
        return capi.getCertificateVersion(binaryCertificate);
    }

    public byte[] sign(byte[] toBeSigned, char[] password) throws Exception {
        if ( FileLog.getVerbose() == FileLog.DEBUG ) {
            FileLog.debug("Signing data:\n" + HexDump.xdump(toBeSigned));
        }
        byte[] data = capi.signMessage(toBeSigned, binaryCertificate);

        if (data == null) {
            int capiError = getLastErrorCode();

            // todo: check errornumber for various conditions and react properly - telling the user when needed etc
            // note that different CSP's can return different codes etc.
            // for now: expect the user to have cancelled
            throw new UserCancel();
        }


        if (data == null)
            throw new IOException("sign was cancelled");
        if ( FileLog.getVerbose() == FileLog.DEBUG ) {
            FileLog.debug("Signature value:\n" + HexDump.xdump(data));
            FileLog.debug("Got error code from microsoft capi wrapper: " + getLastErrorCode());
        }
        return data;
    }

    private int getLastErrorCode() {
        return capi.getLastErrorCode();
    }

    public byte[] digest(byte[] toBeHashed) {
        FileLog.debug("Hashing data:\n" + HexDump.xdump(toBeHashed));

        byte[] data = capi.digest(toBeHashed, "sha1");
        FileLog.debug("Digest value:\n" + HexDump.xdump(data));
        FileLog.debug("Got error code from microsoft capi wrapper: " + getLastErrorCode());

        return data;
    }

    public byte[] getCertificate() {
        return this.binaryCertificate;
    }

    public byte[][] getCertificateChain(char[] password) {
        byte[][] chain = new byte[1][0];
        chain[0] = getCertificate();
        return chain;
    }

    public String getKeyUsage() {
        return "not implemented yet";
    }

    public boolean canSign() {
        KeyUsage ku = getIntendedKeyUsage();
        if (ku == null)
            return true;
        if (ku.includesDigitalSignature())
            return true;
        if (ku.includesNonRepudiation()) {
            return true;
        }

        return false;
    }

    public KeyUsage getIntendedKeyUsage() {
        if (keyUsage == null) {
            keyUsage = getIntendedKeyUsage(getCertificate());
        }
        return keyUsage;
    }

    private KeyUsage getIntendedKeyUsage(byte[] certificate) {

        int i = capi.getKeyUsage(certificate);

        return new KeyUsage((i & 1) != 0,

                (i & 2) != 0,

                (i & 4) != 0,

                (i & 8) != 0,

                (i & 16) != 0,

                (i & 32) != 0,

                (i & 64) != 0);
    }

    public String getStoreName() {
        return "Microsoft Windows";
    }

    // if you want to pass extra info about a certificate (like path to cd-card...)
    public Properties getExtraCertificateProperties() {
        return null;  //To change body of implemented methods use File | Settings | File Templates.
    }
}