/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: Pkcs12CertificateHandler.java,v 1.12 2006/08/12 19:02:35 cara Exp $ */

package org.openoces.opensign.certificate.plugin.pkcs12;

/**
 * This class represents the interface to a pkcs12 certificate
 *
 * @author Carsten Raskgaard  <carsten@raskgaard.dk>
 * @author Preben Valeur  <prv@tdc.dk>
 */

import org.openoces.opensign.certificate.AbstractCertificateHandler;
import org.openoces.opensign.certificate.CertificateInfo;
import org.openoces.opensign.certificate.CopyableItem;
import org.openoces.opensign.certificate.RestoreableItem;
import org.openoces.opensign.certificate.x509.KeyUsage;
import org.openoces.opensign.client.applet.CallBackHandler;
import org.openoces.opensign.client.applet.OS;
import org.openoces.opensign.client.applet.resources.ResourceManager;
import org.openoces.opensign.exceptions.UserCancel;
import org.openoces.opensign.client.applet.bootstrap.FileLog;

import java.awt.*;
import java.io.*;
import java.math.BigInteger;
import java.security.Principal;
import java.security.PrivateKey;
import java.util.Date;
import java.util.Properties;

public class Pkcs12CertificateHandler implements AbstractCertificateHandler, CopyableItem, RestoreableItem {
    private byte [][] chain;
    private CertificateInfo certInfo;
    private String alias;
    private File file;
    private Object keyStore;
    private String[] aliases;
    private CallBackHandler handler;
    private Pkcs12KeyStoreHandler keyStoreHandler;
    private KeyUsage keyusage = null;
    private boolean infoIsAvailable;

    /**
     * @param file
     */
    public Pkcs12CertificateHandler(Pkcs12KeyStoreHandler keyStoreHandler, File file) {
        this.file = file;
        this.keyStoreHandler = keyStoreHandler;
    }

    public void setCallBackHandler(CallBackHandler handler) {
        this.handler = handler;
    }

    /**
     * @param file
     * @param password
     */
    public Pkcs12CertificateHandler(File file, char[] password) throws IOException, Exception {
        this.file = file;
        getKeyStore(password);
        getCertificateChain(password);
    }

    /**
     * If create from same filepath we say they are equal (a second one can replace the first one...)
     *
     * @param o
     * @return
     */
    public boolean equals(Object o) {
        if (o instanceof Pkcs12CertificateHandler) {
            Pkcs12CertificateHandler h = (Pkcs12CertificateHandler) o;
            return (this.file.getAbsolutePath().equals(h.file.getAbsolutePath()));
        } else {
            return false;
        }
    }

    /**
     * @return File
     */
    public File getFile() {
        return file;
    }

    private CertificateInfo getCertInfo(char[] pw) throws IOException, Exception {
        if (certInfo == null) {
            certInfo = OS.getCryptoSupport().getCertInfo(getKeyStore(pw));
        }
        return certInfo;
    }

    public boolean isInfoAvailable() {
        return infoIsAvailable;  //To change body of implemented methods use File | Settings | File Templates.
    }

    /**
     * @return String
     */

    public String getUserfriendlyName() {
        String certificateId = getCertificateID();
        int idx = certificateId.lastIndexOf(OS.getFileSeperator());
        if (idx == -1) {
            return certificateId;
        }
        return certificateId.substring(idx + 1);
    }

    public String getCertificateID() {
        return getFile().toString().replace('\\', '/');
    }


    public String toString() {
        return getFile().toString();
    }

    /**
     * @param password
     * @return KeyStore
     * @throws java.io.IOException
     */

    private synchronized Object getKeyStore(char[] password) throws IOException, Exception {
        if (keyStore == null) {
            keyStore = OS.getCryptoSupport().getKeyStore(file, password);
        }
        return keyStore;
    }

    public Principal getSubjectDN() throws UserCancel, Exception, IOException {
        return getCertInfo(certInfo == null ? getPassword() : null).getSubjectDN();
    }

    public Principal getIssuerDN() throws UserCancel, Exception, IOException {
        return getCertInfo(certInfo == null ? getPassword() : null).getIssuerDN();
    }

    ;

    public BigInteger getSerialNumber() throws UserCancel, Exception {
        return getCertInfo(certInfo == null ? getPassword() : null).getSerialNumber();
    }

    public Date getNotBefore() throws UserCancel, Exception {
        return getCertInfo(certInfo == null ? getPassword() : null).getNotBefore();
    }

    public Date getNotAfter() throws UserCancel, Exception {
        return getCertInfo(certInfo == null ? getPassword() : null).getNotAfter();
    }

    public int getVersion() throws UserCancel, Exception {
        return getCertInfo(certInfo == null ? getPassword() : null).getVersion();
    }

    public byte[] sign(byte[] toBeSigned, char[] password) throws IOException, Exception {
        if (password == null || password.length == 0) {
            password = getPassword();
        }
        if (canSign(password)) {
            PrivateKey pk = OS.getCryptoSupport().getPrivateKey(getKeyStore(password), password);
            return OS.getCryptoSupport().sign(toBeSigned, pk);
        } else
            throw new Exception("certificate not signable");
    }

    public byte[] digest(byte[] toBeHashed) throws Exception {
        return OS.getCryptoSupport().sha1(toBeHashed);
    }

    public byte[] getCertificate() throws UserCancel, Exception, IOException {
        return getCertInfo(certInfo == null ? getPassword() : null).getCertificate();
    }

    /**
     *
     */
    public byte[][] getCertificateChain(char[] password) throws IOException, Exception {
        infoIsAvailable = true;
        if (chain == null) {
            chain = OS.getCryptoSupport().getCertificateChain(getKeyStore(password));
        }
        return chain;
    }

    public void deleteCertificate() {
        File f = this.getFile();
        FileLog.debug("deleting file:" + f);
        keyStoreHandler.getRestrictedAction().delete(f);
    }

    public String getKeyUsage() throws Exception {
        return getCertInfo(certInfo == null ? getPassword() : null).getKeyUsage();
    }

    private boolean canSign(char[] pw) {
        try {
            return getCertInfo(certInfo == null ? getPassword() : null).canSign();
        } catch (Exception e) {
            return false;
        }
    }

    public boolean canSign() throws IOException, UserCancel {
        return canSign(certInfo == null ? getPassword() : null);
    }

    private char[] getPassword() throws UserCancel {
        while (true) {
            char [] pw = handler.promptPassword(ResourceManager.getString("DLG_PASSWORD_HEADER_PREFIX") + " " + file).toCharArray();
            try {
                getCertInfo(pw);
                return pw;
            } catch (Exception e) {
                FileLog.warn("Invalid password entered");
            }
        }
    }

    public KeyUsage getIntendedKeyUsage() throws Exception, IOException {
        if (keyusage == null) {
            keyusage = getCertInfo(certInfo == null ? getPassword() : null).getIntendedKeyUsage();
        }
        return keyusage;
    }

    // copy the item from the default place to a user selected place
    public void backup(Frame parent) throws IOException {

        FileDialog dlg = keyStoreHandler.getRestrictedAction().getFileDialog(parent, ResourceManager.getString("BACKUP_KEYFILE"), FileDialog.SAVE);
        dlg.setFilenameFilter(new FilenameFilter() {
            public boolean accept(File file, String n) {
                return n.toLowerCase().endsWith("pfx") || n.toLowerCase().endsWith("p12");
            }
        });
        dlg.setDirectory(keyStoreHandler.getLastDirectoryName().toString());
        dlg.setModal(true);
        dlg.setFile(getFile().toString());
        keyStoreHandler.getRestrictedAction().showFileDialog(dlg);

        if (dlg.getFile() == null) {
            FileLog.debug("dialog cancelled");
            return;
        }

        File src = getFile();
        File dest = new File(dlg.getDirectory() + OS.getFileSeperator() + dlg.getFile());

        FileLog.debug("Copying pkcs12 from " + src + " to " + dest);

        if (dest != null) {
            keyStoreHandler.copy(src, dest);
        }
    }


    // delete from the default place
    public void delete(Frame parent) {
        deleteCertificate();
    }

    // copy the item from another place to the default place
    public void restore(Frame parent) throws IOException {
        keyStoreHandler.importPKCS12(parent);
    }

    private void prepareKeyDir() throws IOException {
        keyStoreHandler.prepareKeyDir();
    }

    public String getStoreName() {
        // todo: own Resources? ResourceManager?
        return ResourceManager.getString("DLG_CERTIFICATEDETAILS_LABEL_CERTIFICATESTORE_PKCS12") + " (" + this.getFile().toString() + ")";
    }

    // if you want to pass extra info about a certificate (like path to cd-card...)
    public Properties getExtraCertificateProperties() {
        return null;  //To change body of implemented methods use File | Settings | File Templates.
    }

    public boolean supportsDelete() {
        return true;
    }

}