/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: SunTruststoreFile.java,v 1.5 2006/08/12 19:02:35 cara Exp $ */

package org.openoces.opensign.certificate.support.suncrypto;

import org.openoces.opensign.client.applet.bootstrap.interfaces.RestrictedAction;
import org.openoces.opensign.client.applet.bootstrap.FileLog;
import org.openoces.opensign.utils.HexDump;

import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * This class represents the ssl trust store used for ssl communications
 *
 * @author Preben Valeur  <prv@tdc.dk>
 */
public class SunTruststoreFile {

    private static String OCES_INSTALL_DIRECTORY = ".oces";
    private static String OPENSIGN_INSTALL_DIRECTORY = "opensign";
    private static String TRUSTSTORE_FILENAME = "truststore.jks";
    private static final String TRUSTSTORE_PASSWORD = "default";

    private static final byte[] DIGESTVALUE = {(byte) 0x31, (byte) 0xdf, (byte) 0xd9, (byte) 0x7a, (byte) 0xec, (byte) 0xd0, (byte) 0x6d, (byte) 0x9c,
                                               (byte) 0xcb, (byte) 0x23, (byte) 0xd3, (byte) 0x13, (byte) 0x41, (byte) 0xbb, (byte) 0x38, (byte) 0x54,
                                               (byte) 0x47, (byte) 0xd3, (byte) 0x8c, (byte) 0x4a};

    private RestrictedAction ra;

    public SunTruststoreFile(RestrictedAction ra) {
        this.ra = ra;
    }

    private File getInstallDirectory() {
        String relativePath = OCES_INSTALL_DIRECTORY + File.separator + OPENSIGN_INSTALL_DIRECTORY;
        return new File(ra.getUserHome() + File.separator + relativePath);
    }

    private File getAbsoluteFile() {
        return new File(getInstallDirectory(), TRUSTSTORE_FILENAME);
    }

    public boolean isValid() {
        FileInputStream ios = null;
        MessageDigest sha;

        try {
            byte[] data;
            byte[] hash;
            int readBytes;

            data = new byte[256];

            sha = MessageDigest.getInstance("SHA-1");
            ios = ra.getFileInputStream(getAbsoluteFile());

            while ((readBytes = ios.read(data)) != -1) {
                sha.update(data, 0, readBytes);
            }
            ios.close();
            hash = sha.digest();

            if (hash.length != DIGESTVALUE.length) {
                FileLog.warn("Calculated and stored digest have different lengths");
                return false;
            }

            for (int i = 0; i < hash.length; i++) {
                byte b = hash[i];
                byte c = DIGESTVALUE[i];
                if (b != c) {
                    FileLog.warn("Calculated and stored digest do not match at position " + i);
                    FileLog.warn("Calculated hash:\n" + HexDump.xdump(hash));
                    FileLog.warn("Stored hash:\n" + HexDump.xdump(DIGESTVALUE));
                    return false;
                }
            }
            FileLog.debug("Truststore verification was successful");
            return true;

        } catch (FileNotFoundException e) {
            FileLog.warn("Truststore to be verified not found");
        } catch (IOException e) {
            FileLog.warn("IO error occured while verifying truststore integrity");
        } catch (NoSuchAlgorithmException e) {
            FileLog.warn("Missing algorithm while verifying truststore integrity");
        } catch (NullPointerException e) {
            FileLog.warn("Got a null pointer exception");
        }
        return false;
    }

    public boolean download() {
        URLConnection urlCon = null;
        InputStream fis = null;

        try {
            FileLog.debug("Base url is " + ra.getBaseUrl());
            urlCon = new URL(ra.getBaseUrl() + "/" + TRUSTSTORE_FILENAME).openConnection();
            ra.connect(urlCon);
            FileLog.debug("connected to " + urlCon);
            fis = urlCon.getInputStream();
        } catch (IOException e) {
            FileLog.warn("failed to open url connection");
            return false;
        }

        if (urlCon == null) {
            FileLog.warn("truststore not available");
            return false;
        }

        FileLog.debug("Downloading truststore from " + urlCon);

        try {

            File f = this.getAbsoluteFile();
            /* FIXME: check that the directory exists? */

            FileOutputStream fos = ra.getFileOutputStream(f);
            int ch;

            while ((ch = fis.read()) != -1) {
                fos.write(ch);
            }

            fis.close();
            fos.close();

        } catch (IOException e) {
            FileLog.error("An IO exception occured");
            return false;
        }
        return true;
    }

    public void useAsTruststore() {
        FileLog.debug("using truststore password: " + TRUSTSTORE_PASSWORD);
        FileLog.debug("using truststore located at: " + this.getAbsoluteFile().toString());
        ra.setProperty("javax.net.ssl.trustStorePassword", TRUSTSTORE_PASSWORD);
        ra.setProperty("javax.net.ssl.trustStore", this.getAbsoluteFile().toString());
    }

    public InputStream getTrustStoreInputStream() {
        return ra.getFileInputStream(getAbsoluteFile());
    }

    public String getTrustStorePassword() {
        return TRUSTSTORE_PASSWORD;
    }
}
