/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: AttachmentImpl.java,v 1.9 2006/10/22 08:48:14 cara Exp $ */

package org.openoces.opensign.client.applet.attach;

/**
 * This class implements a generic attachment
 *
 * @author Preben Rosendal Valeur  <prv@tdc.dk>
 * @author Carsten Raskgaard  <carsten@raskgaard.dk>
 */

import org.openoces.opensign.xml.nanoxml.XMLElement;
import org.openoces.opensign.client.applet.attach.resources.AttachmentResourceManager;
import org.openoces.opensign.client.applet.dialogs.DlgInformation;
import org.openoces.opensign.client.applet.bootstrap.FileLog;
import org.openoces.opensign.utils.Base64;
import org.openoces.opensign.utils.HexDump;
import org.openoces.opensign.appletsupport.Attachment;
import org.openoces.opensign.appletsupport.MimeTypeValidator;

import java.util.Vector;
import java.util.Observable;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.*;
import java.io.*;

public class AttachmentImpl extends Observable implements ActionListener, Attachment {
    private String title;
    private String path;
    private String mimeType = "undefined";
    private long size;
    private String hashAlg = "sha1";
    private boolean optional;
    private boolean hasSeen;
    private Sha1Checksum hashValue;
    private Sha1Checksum localHashValue;

    private boolean checksumVerified;
    private AttachmentServer server;
    private Frame dialogOwner;

    AttachmentImpl(XMLElement param, AttachmentServer server) {
        this.server = server;
        Vector children = param.getChildren();
        for (int i = 0; i < children.size(); i++) {
            XMLElement child = (XMLElement) children.elementAt(i);
            String name = child.getName();
            if (name.equals("title")) {
                title = child.getContent();
            } else if (name.equals("path")) {
                path = child.getContent();
            } else if (name.equals("size")) {
                size = Long.parseLong(child.getContent());
            } else if (name.equals("hashValue")) {
                byte[] bs = Base64.decode(child.getContent().getBytes());
                FileLog.debug("read checksum: " + HexDump.xdump(bs));
                hashValue = new Sha1Checksum(bs);
            } else if (name.equals("hashAlg")) {
                hashAlg = child.getContent();
            } else if (name.equals("mimeType")) {
                String unvalidatedMimeType = child.getContent();
                if (MimeTypeValidator.validate(unvalidatedMimeType)) {
                    mimeType = unvalidatedMimeType;
                } else {
                    FileLog.warn("invalid mime type for an attachment: " + unvalidatedMimeType + ". Using default mime-type.");
                    mimeType = "text/plain";
                }
            } else if (name.equals("optional")) {
                optional = true;
            }
        }
    }

    public String toString() {
        return "Attachment:\n" +
                "\ttitle: " + getTitle() +
                "\tpath: " + getPath() +
                "\tmimeType: " + getMimeType() +
                "\tsize: " + getSize() +
                "\thashAlg: " + getHashAlg() +
                "\thashValue: " + getHashValue();
    }

    public String getTitle() {
        return title;
    }

    public boolean isViewable() {
        return AttachmentViewerManager.hasViewer(mimeType);
    }

    public void setDialogOwner(Frame owner) {
        this.dialogOwner = owner;
    }

    public void view() {
        if (!isViewable()) {
            FileLog.warn("Tried to view unviewable attachment");
            return;
        }

        AttachmentViewer viewer = AttachmentViewerManager.getViewer(mimeType);
        FileLog.debug("AttachmentViewer for mimetype: " + mimeType + " : " + viewer);
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        OutputStream toBuf = new BufferedOutputStream(bos);
        if (!server.getContent(this, toBuf)) {
            FileLog.warn("unable to download attachment: " + this.getPath());
            String ok = AttachmentResourceManager.getString("OK");
            DlgInformation dlgInfo = new DlgInformation(server.getRestrictedAction(), dialogOwner, "fixme", "fixme: unable to download attachment", ok);
            dlgInfo.show();
            return;
        }
        viewer.view(dialogOwner, getTitle(), bos.toByteArray());
        setHasSeen();
    }

    private void setHasSeen() {
        hasSeen = true;
        this.setChanged();
        this.notifyObservers();
    }

    public void save() {
        // openfiledialog and then:
        String filePath = server.getAttachmentFilePath(getPath());

        if (filePath == null) {
            FileLog.fatal("filePath is null");
            // cancelled
            return;
        }
        File f = new File(filePath);

        OutputStream toFile = new BufferedOutputStream(server.getRestrictedAction().getFileOutputStream(f));
        if (server.getContent(this, toFile)) {
            setHasSeen();
        } else {
            if (server.getRestrictedAction().exists(f)) {
                server.getRestrictedAction().delete(f);
            }
            String ok = AttachmentResourceManager.getString("OK");
            DlgInformation dlgInfo = new DlgInformation(server.getRestrictedAction(), dialogOwner,
                    AttachmentResourceManager.getString("ATTACHMENT_SAVE_FAILED_HEADER"),
                    AttachmentResourceManager.getString("ATTACHMENT_SAVE_FAILED_BODY"), ok);
            dlgInfo.show();
        }
    }

    public String getPath() {
        return path;
    }

    public boolean isHasSeen() {
        return hasSeen;
    }

    /**
     * Returns a textual representation of the hash value of the attachment as described by the server.
     *
     * @return the textual representation of the hash of the attachment
     */
    public String getHashValue() {
        return new String(Base64.encode(hashValue.getSha1Value()));
    }

    public Checksum getLocalChecksum() {
        return localHashValue;
    }

    public void setLocalChecksum(Sha1Checksum v) {
        localHashValue = v;
    }

    public Checksum getChecksum() {
        /* return a new object to avoid unintended manipulation of the value */
        return new Sha1Checksum(hashValue.getSha1Value());
    }

    public String getHashAlg() {
        return hashAlg;
    }

    public String getMimeType() {
        return mimeType;
    }

    public byte[] getContents() {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        if (!server.getContent(this, bos)) {
            FileLog.warn("unable to download attachment: " + this.getPath());
            String ok = AttachmentResourceManager.getString("OK");
            DlgInformation dlgInfo = new DlgInformation(server.getRestrictedAction(), dialogOwner, "fixme", "fixme: unable to download attachment", ok);
            dlgInfo.show();
            return null;
        }
        return bos.toByteArray();
    }

    public long getSize() {
        return size;
    }

    public boolean isOptional() {
        return optional;
    }

    public void actionPerformed(ActionEvent e) {
        // view or save
        if (e.getActionCommand().equals(AttachmentResourceManager.getString("VIEW"))) {
            view();
        } else if (e.getActionCommand().equals(AttachmentResourceManager.getString("SAVE"))) {
            save();
        }
    }

    public boolean getChecksumVerified() {
        return checksumVerified;
    }

    public void setChecksumVerified() {
        checksumVerified = true;
    }
}
