/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: AttachmentServer.java,v 1.13 2006/10/22 08:41:04 cara Exp $ */

package org.openoces.opensign.client.applet.attach;

/**
 * This class represents a server serving attachments
 *
 * @author Preben Rosendal Valeur  <prv@tdc.dk>
 * @author Carsten Raskgaard  <carsten@raskgaard.dk>
 */

import org.openoces.opensign.client.applet.bootstrap.FileLog;
import org.openoces.opensign.utils.Base64;
import org.openoces.opensign.client.applet.bootstrap.interfaces.RestrictedAction;
import org.openoces.opensign.client.applet.OcesAppletUtils;
import org.openoces.opensign.client.applet.CallBackHandler;
import org.openoces.opensign.client.applet.attach.resources.AttachmentResourceManager;
import org.openoces.opensign.client.applet.dialogs.DlgConfirm;
import org.openoces.opensign.client.applet.dialogs.DlgInformation;
import org.openoces.opensign.client.applet.resources.ResourceManager;
import org.openoces.opensign.xml.nanoxml.XMLElement;
import org.openoces.opensign.appletsupport.Attachment;

import java.net.*;
import java.io.*;
import java.util.Vector;
import java.applet.Applet;
import java.awt.*;
import java.security.NoSuchAlgorithmException;

public class AttachmentServer {
    private RestrictedAction ra;
    private Applet applet;
    private AttachmentImpl[] attachments;

    /**
     * For the benefit of AttachmentImpl
     *
     * @return the RestrictedAction
     */
    RestrictedAction getRestrictedAction() {
        return ra;
    }

    public AttachmentServer(RestrictedAction ra, Applet applet, String attachmentParam) {
        this.applet = applet;
        this.ra = ra;
        String attachmentXml = null;
        byte[] ba = Base64.decode(attachmentParam.getBytes());
        try {
            attachmentXml = new String(ba, "UTF8");
        } catch (UnsupportedEncodingException e) {
            FileLog.fatal(e.toString());
        }

        FileLog.info("xml:\r\n\"" + attachmentXml.substring(0, Math.min(300, attachmentXml.length())) + "\"");
        XMLElement doc = new XMLElement();
        doc.parseString(attachmentXml);
        Vector children = doc.getChildren();
        attachments = new AttachmentImpl[children.size()];
        for (int i = 0; i < children.size(); i++) {
            XMLElement child = (XMLElement) children.elementAt(i);
            attachments[i] = new AttachmentImpl(child, this);
        }
    }

    public Attachment[] getAttachments() {
        return attachments;
    }

    // sample mime types:
    // from http://www.w3schools.com/media/media_mimeref.asp
    // application/msword       doc
    // application/pdf          pdf
    // application/postscript   ps
    // application/rtf          rtf
    // image/bmp                bmp
    // image/gif                gif
    // image/jpeg               jpg/jpe/jpeg
    // image/tiff               tif/tiff
    // text/html                htm/html
    // text/plain               txt/...
    //

    /**
     * Return true if succesfully saved file
     * AttachmentImpl status is updated with state of checksum comparison
     */
    private boolean copyStream(AttachmentImpl a, InputStream in, OutputStream out) throws IOException {
        FileLog.debug("copyStream called a:" + a + " in: " + in + " out: " + out);
        Sha1ChecksummedInputStream checksummedStream;
        synchronized (in) {
            synchronized (out) {
                try {
                    checksummedStream = new Sha1ChecksummedInputStream(in, a.getChecksum());
                } catch (NoSuchAlgorithmException e) {
                    return false;
                }
                byte[] buffer = new byte[64 * 1024];
                int c;
                // int offset = 0; // ?? there seems to be a funny char when utf8 which can
                // be removed by setting offset to 3 - but then other formats doesn't work
                // so maybe check the first utf8-char where needed?
                while ((c = checksummedStream.read(buffer, 0, 64 * 1024)) > -1) {
                    FileLog.debug("writing " + c + " bytes");
                    out.write(buffer, 0, c);
                }
            }
        }

        try {
            if (a.isHasSeen() && !checksummedStream.getCalculatedChecksum().matches(a.getLocalChecksum())) {
                // the attachment is being modified while the applet is running. We do not allow this.
                // We throw an exception and do not update the local hash of the attachment
                FileLog.debug("Checksum has changed between two downloads of the same attachment");
                throw new Exception();
            }

            a.setLocalChecksum(checksummedStream.getCalculatedChecksum());

            FileLog.debug("About to verify attachment checksum");
            if (checksummedStream.hasChecksumValid()) {
                a.setChecksumVerified();
                return true;
            } else {
                FileLog.warn("Invalid checksum");
                FileLog.debug("Calculated checksum (base64 encoded): " + new String(Base64.encode(checksummedStream.getCalculatedChecksum().getSha1Value())));
                FileLog.debug("Expected checksum (base64 encoded): " + new String(Base64.encode(((Sha1Checksum) a.getChecksum()).getSha1Value())));
                return false;
            }
        } catch (Exception e) {
            FileLog.warn("Exception checking checksum " + e.getMessage());
            throw new IOException();
        }
    }

    public String getAttachmentFilePath(String serverPath) {
        String defaultFileName = serverPath.substring(serverPath.lastIndexOf("/") + 1);
        FileDialog dlg = ra.getFileDialog(((CallBackHandler) applet).getFrame(), AttachmentResourceManager.getString("SAVE_ATTACHMENT"), FileDialog.SAVE);
        dlg.setDirectory(ra.getUserHome());
        dlg.setFile(defaultFileName);
        dlg.setModal(true);
        ra.showFileDialog(dlg);
        String attachmentFileName = dlg.getFile();
        String attachmentDirName = dlg.getDirectory();
        if (dlg.getFile() == null) {
            FileLog.debug("dialog cancel");
            return null;
        }

        String fPath = attachmentDirName + attachmentFileName;

        File f = new File(fPath);

        if (getRestrictedAction().exists(f)) {
            String message = AttachmentResourceManager.getString("CONFIRM_OVERWRITE_FILE");
            DlgConfirm dlgConfirm = new DlgConfirm(getRestrictedAction(), (((CallBackHandler) applet).getFrame()), message, message);
            dlgConfirm.show();
            if (dlgConfirm.isCancelled()) {
                return null;
            }
        }

        return fPath;
    }

    boolean getContent(AttachmentImpl a, OutputStream os) {
        // get the url of the applet
        String docBase = OcesAppletUtils.getDocBase();
        String realUrlString = docBase + a.getPath();
        boolean streamOk;
        try {
            // put it into a file:
            URL realUrl = new URL(realUrlString);
            URLConnection con = realUrl.openConnection();
//            InputStream fromServer = ra.getInputStream(con);
            // instead of the line above we do this:
            ra.connect(con);
            InputStream fromServer = con.getInputStream();
            // end do this
            if (fromServer == null) {
                FileLog.error("Unable to download attachment from: "+a.getPath() + "( full url: " + realUrl + ")");
                return false;
            }
            FileLog.debug("calling copyStream 1");
            streamOk = this.copyStream(a, fromServer, os);
            if (!streamOk) {
                FileLog.error("Invalid checksum detected. Attachment info: "+a.getPath() + "( full url: " + realUrl + ")");
            }
            FileLog.debug("copyStream done");
            fromServer.close();
            os.flush();
            os.close();
        } catch (MalformedURLException e) {
            FileLog.error("Error showing attachment " + e.getMessage());
            return false;
        } catch (IOException e) {
            FileLog.error("Error showing attachment " + e.getMessage());
            return false;
        }
        return streamOk;
    }
}
