/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: MultiClassLoader.java,v 1.4 2006/04/02 08:52:36 cara Exp $ */

package org.openoces.opensign.client.applet.bootstrap.classloader;

import org.openoces.opensign.client.applet.bootstrap.FileLog;

import java.util.Hashtable;

/**
 * A simple test class loader capable of loading from
 * multiple sources, such as local files or a URL. Must be
 * subclassed and the abstract method loadClassBytes()
 * implemented to provide the preferred source.
 * <p/>
 * This class is derived from an article by Chuck McManis
 * http://www.javaworld.com/javaworld/jw-10-1996/indepth.src.html
 * with large modifications.
 *
 * @author Jack Harich
 */
public abstract class MultiClassLoader extends ClassLoader {
    private static String trustedClass;

    /* the logging is rather nosisy, so it is nice to be able to disable it */
    private boolean doLog = false;

    // is necessary when you have a class backing a native lib - it needs to be part of the signed code for MS
    // todo: change so it supports more than one class
    public static void addTrustedClass(String className) {
        trustedClass = className;
    }

    private boolean isTrusted(String className) {
        return (trustedClass != null && className.indexOf(trustedClass) >= 0);
    }

//---------- Fields --------------------------------------
    private Hashtable classes = new Hashtable();
    private char classNameReplacementChar;

    protected boolean monitorOn = true;
    protected boolean sourceMonitorOn = true;
    protected ClassLoader parent;

//---------- Initialization ------------------------------
    protected MultiClassLoader() {

    }

    protected MultiClassLoader(ClassLoader parent) {
        this.parent = parent;
    }
//---------- Superclass Overrides ------------------------
    /**
     * This is a simple version for external clients since they
     * will always want the class resolved before it is returned
     * to them.
     */
    public Class loadClass(String className) throws ClassNotFoundException {
        return (loadClass(className, true));
    }
//---------- Abstract Implementation ---------------------
    public synchronized Class loadClass(String className,
                                        boolean resolveIt) throws ClassNotFoundException {

        Class result = null;
        byte[] classBytes = null;
        if (doLog) FileLog.debug(">> MultiClassLoader.loadClass(" + className + ", " + resolveIt + ")" + this);

        if (doLog) FileLog.debug(">> checking if it is allready loaded");
        result = findLoadedClass(className);
        if (result != null) {
            if (doLog) FileLog.debug(">> MultiClassLoader.loadClass(" + className + " is allready loaded <<<<<");
            return result;
        }
        //----- Check our local cache of classes
        result = (Class) classes.get(className);
        if (result != null) {
            if (doLog) FileLog.debug(">> returning cached result.");
            return result;
        }
        //----- Check with the primordial class loader
//    if ((className.indexOf("openoces") == -1) || className.indexOf("bootstrap") != -1){
        try {
            result = super.findSystemClass(className);
            if (doLog) FileLog.debug(">> returning system class (in CLASSPATH).");
            return result;
        } catch (ClassNotFoundException e) {
            if (doLog) FileLog.debug(">> Not a system class.");
        }
//    }
        //----- Try to load it from preferred source
        // Note loadClassBytes() is an abstract method

        if (isTrusted(className)) {
            // get the trusted version from the parent if possible!
            try {
                result = parent.loadClass(className);
                if (doLog) FileLog.debug(">> *************** Special *********** MultiClassLoader.loadClass(" + className + " parent loaded <<<<<");
                return result;
            } catch (ClassNotFoundException e) {
                if (doLog) FileLog.debug(">> *************** Special ***********  parent couldn't load - continuing");
            }
        }
        classBytes = loadClassBytes(className);
        if (parent != null && classBytes == null) {
            try {
                result = parent.loadClass(className);
                if (doLog) FileLog.debug(">> MultiClassLoader.loadClass(" + className + " parent loaded <<<<<");
                return result;
            } catch (ClassNotFoundException e) {
                if (doLog) FileLog.debug(">> parent couldn't load - continuing");
            }
        }
        if (classBytes == null) {
            if (resolveIt) {
                if (doLog) FileLog.debug(">>> null AND resolveIt, we throw ClassNotFoundException !!");
                throw new ClassNotFoundException();
            } else {
                if (false) {
                    if (doLog) FileLog.debug(">>> null from loadClassBytes() but we don't throw ClassNotFoundException !!");
                    return null; // todo: find out if this is ok!
                } else {
                    if (doLog) FileLog.debug(">>> null but not resolveIt, we throw ClassNotFoundException !! (in this version..)");
                    throw new ClassNotFoundException();
                }
            }
        }

        //----- Define it (parse the class file)
        result = defineClass(classBytes, 0, classBytes.length);
        if (result == null) {
            throw new ClassFormatError();
        }

        //----- Resolve if necessary
        if (resolveIt) resolveClass(result);

        // Done
        classes.put(className, result);
        if (doLog) FileLog.debug(">> Returning newly loaded class.");
        return result;
    }
//---------- Public Methods ------------------------------
    /**
     * This optional call allows a class name such as
     * "COM.test.Hello" to be changed to "COM_test_Hello",
     * which is useful for storing classes from different
     * packages in the same retrival directory.
     * In the above example the char would be '_'.
     */
    public void setClassNameReplacementChar(char replacement) {
        classNameReplacementChar = replacement;
    }
//---------- Protected Methods ---------------------------
    protected abstract byte[] loadClassBytes(String className);

    protected String formatClassName(String className) {
        if (classNameReplacementChar == '\u0000') {
            // '/' is used to map the package to the path
            return className.replace('.', '/') + ".class";
        } else {
            // Replace '.' with custom char, such as '_'
            return className.replace('.',
                    classNameReplacementChar) + ".class";
        }
    }
} // End class
