/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: ZipClassLoader.java,v 1.7 2006/04/02 08:52:36 cara Exp $ */

package org.openoces.opensign.client.applet.bootstrap.classloader;

import org.openoces.opensign.client.applet.bootstrap.DownloadItem;
import org.openoces.opensign.client.applet.bootstrap.FileLog;
import org.openoces.opensign.client.applet.bootstrap.interfaces.RestrictedAction;
import org.openoces.opensign.client.Configuration;

import java.io.*;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.util.Hashtable;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * This class provides functionality for loading classes from a zip files. This is
 * used by the plugin mechanism.
 *
 * @author Preben Valeur  <prv@tdc.dk>
 */
public class ZipClassLoader extends MultiClassLoader {

    private static final int CLASS_BUFFER_SIZE = 64 * 1024;//todo: max size for a class - check
    // todo: set to false when ready!

    private static RestrictedAction ra;

    public static void setRestrictedAction(RestrictedAction res) {
        // only when beta - for stripping jars
        ra = res;
    }

    private DownloadItem item;

    public ZipClassLoader(DownloadItem item) throws IOException, NoSuchAlgorithmException {
        if (!verifyAndLoad(item)) {
            throw new IOException("checksums don't match");
        }
    }

    public ZipClassLoader(DownloadItem item, ClassLoader parent) throws IOException, NoSuchAlgorithmException {
        super(parent);
        if (!verifyAndLoad(item)) {
            throw new IOException("checksums don't match");
        }
    }


    /**
     * Checks if there is a zip-file in cache of the right name with the correct checkSum
     * If not: gets it from the url and puts it into cache.
     * Open it and scans through to check checksum.
     * If not ok: deletes it and fail,
     *
     * @param item
     * @return if it succeeded
     */
    private boolean verifyAndLoad(DownloadItem item) throws IOException, NoSuchAlgorithmException {
        this.item = item;

        Hashtable cache = new Hashtable();
        boolean loaded = false;

        if (!item.cachedExists()) {
            FileLog.debug("no cached version exists of " + item.getAlias());
            item.download();
            if (!item.cachedExists()) {
                FileLog.error("Unable to create local version of: " + item.getAlias());
                return false;
            }
        }

        loaded = loadPlugin(item.getAsStream(), item, cache);
        if (!loaded) {
            FileLog.error("Cached version of " + item.getAlias() + " failed to load");
            item.download();
            loaded = loadPlugin(item.getAsStream(), item, cache);
            if (loaded) {
                FileLog.debug("Loaded plugin " + item.getAlias() + " correctly");
            } else {
                FileLog.debug("Failed to load plugin " + item.getAlias() + " in the second try also");
            }
        }

        if (loaded) {
            setCache(cache);
            return true;
        }
        return false;
    }

    private boolean loadPlugin(InputStream fis, DownloadItem item, Hashtable cache) throws IOException, NoSuchAlgorithmException {
        Sha1ChecksummedInputStream checksummedStream = new Sha1ChecksummedInputStream(fis, item.getChecksum());

        byte buf[] = new byte[CLASS_BUFFER_SIZE];
        ByteArrayOutputStream bos = new ByteArrayOutputStream();

        int c;
        while ((c = checksummedStream.read(buf, 0, 64 * 1024)) > -1) {
            bos.write(buf, 0, c);
            item.noteProgress(c);
        }
        item.noteDone();

        try {
            if (!Configuration.doChecksumPlugins()) FileLog.info("ZipClassLoader: checksumming of plugins disabled. This is dangerous");
            if (!Configuration.doChecksumPlugins() || checksummedStream.hasChecksumValid()) {
                ZipInputStream zis = new ZipInputStream(new BufferedInputStream(new ByteArrayInputStream(bos.toByteArray())));
                ZipEntry entry;

                while ((entry = zis.getNextEntry()) != null) {
                    FileLog.debug("Extracting: " + entry);
                    if (entry.isDirectory()) {
                        FileLog.debug(" it is a directory, skipping it");
                        continue;
                    }
                    int count;
                    int datasize = (int) entry.getSize();
                    byte[] data = null;
                    int pos = 0;
                    if (datasize != -1) {
                        // we know the size
                        FileLog.debug("known datasize :" + datasize);
                        data = new byte[datasize];
                        while ((count = zis.read(data, pos, data.length - pos)) != -1) {

                            FileLog.debug("known size: " + datasize + " read " + count + " bytes total: " + (pos + count));
                            pos += count;
                            if (pos == datasize) {
                                FileLog.debug("All read");
                                break;
                            }
                        }
                    } else {
                        // unknown size
                        FileLog.debug("unknown datasize");
                        while ((count = zis.read(buf, pos, buf.length - pos)) != -1) {
                            item.noteProgress(count);
                            FileLog.debug("unknown - read " + count + " bytes total: " + (pos + count));
                            pos += count;
                            if (pos == datasize) {
                                FileLog.debug("All read");
                                break;
                            }
                        }
                        data = new byte[pos];
                        System.arraycopy(buf, 0, data, 0, pos);
                    }
                    FileLog.debug("putting in cache");
                    cache.put(entry.getName(), data);
                }
                zis.close();

                FileLog.debug("Finished loading plugin.");
                return true;
            } else {
                FileLog.error("Invalid checksum occurred");
                return false;
            }
        } catch (Exception e) {
            e.printStackTrace();
            FileLog.error("An exception occured while verifying the checksum: " + e);
        }
        return false;
    }

    private Hashtable cachedEntries;

    private void setCache(Hashtable entries) {
        this.cachedEntries = entries;
    }

    private void saveClassBytes(File dir, String classPath, byte[] classBytes) {
        try {
            // write the files to the disk
            String fullPath = dir + File.separator + classPath + ".class";
            fullPath = fullPath.replace('/', File.separatorChar);
            File f = new File(fullPath);
            String parent = f.getParent();
            if (parent != null) {
                File parentDir = new File(parent);
                if (!ra.exists(parentDir)) {
                    if (!ra.mkdirs(parentDir)) {
                        FileLog.error("Couldn't make parentDir: " + parentDir);
                        return;
                    }
                }
            }
//            File parentDir = f.getParentFile(); not 1.1
            FileOutputStream fos = ra.getFileOutputStream(f);
            if (fos == null) {
                FileLog.error("!!! FOS is NULL");
                return;
            }
            fos.write(classBytes);
            fos.close();
        } catch (IOException e) {
            FileLog.error("IOException saving loaded classfiles: " + e.getMessage());
        }
    }

    protected byte[] loadClassBytes(String className) {
        // just get it from our local cache!
        FileLog.debug("loading " + className);
        className = className.replace('.', '/');
        if (item.getDirForUsedClasses() != null) {
            // save it
            byte[] classBytes = (byte[]) cachedEntries.get(className + ".class");
            if (classBytes != null) {
                //todo: disabled for now - gets nullpointerexception -put back later!
//                saveClassBytes(item.getDirForUsedClasses(), className, classBytes);
            }
            return classBytes;
        } else {
            return (byte[]) cachedEntries.get(className + ".class");
        }
    }

    protected URL findResource(String name) {
        FileLog.debug("### findResource(" + name + ") called  - but NOT IMPLEMENTED YET! (returning null)");
//        return new URL("jar", "", "file:" + mscapiJarPath.getAbsolutePath() + "!/");
        return null;
    }

    public String toString() {
        return "ZipClassLoader of " + this.item.getName() + " " + super.toString();
    }
}
