/*
    Copyright 2006 IT Practice A/S
    Copyright 2006 TDC Totalløsninger A/S
    Copyright 2006 Jens Bo Friis
    Copyright 2006 Preben Rosendal Valeur
    Copyright 2006 Carsten Raskgaard


    This file is part of OpenSign.

    OpenSign is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    OpenSign is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OpenOcesAPI; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Note to developers:
    If you add code to this file, please take a minute to add an additional
    copyright statement above and an additional
    @author statement below.
*/

/* $Id: BasicSignatureGenerator.java,v 1.12 2006/10/21 22:05:11 cara Exp $ */

package org.openoces.opensign.xml.xmldsig;

import org.openoces.opensign.certificate.AbstractCertificateHandler;
import org.openoces.opensign.exceptions.InputDataError;
import org.openoces.opensign.utils.Base64;
import org.openoces.opensign.client.applet.bootstrap.FileLog;
import org.openoces.opensign.appletsupport.Attachment;
import org.openoces.opensign.appletsupport.MimeTypeValidator;
import org.openoces.opensign.client.applet.bootstrap.HexDump;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Enumeration;
import java.util.Properties;

/**
 * This class implements a xmldsig generator for the standard xmldsig format
 *
 * @author Carsten Raskgaard  <carsten@raskgaard.dk>
 */

public class BasicSignatureGenerator implements SignatureGenerator {

    public static final String NS_OPENOCES_SIGNATURE = "http://www.openoces.org/2006/07/signature#";
    public static final String NS_W3_XMLDSIG = "http://www.w3.org/2000/09/xmldsig#";

    public StringBuffer logon(AbstractCertificateHandler certificate, char[] password, Properties visibleProps, Properties invisibleProps) throws Exception {
        invisibleProps.put(ACTION, ACTION_LOGON);

        return getSignature(certificate, password, null, visibleProps, invisibleProps, null);
    }

    public StringBuffer sign(AbstractCertificateHandler certificate, char[] password, StringBuffer signText, Properties visibleProps, Properties invisibleProps, Attachment[] attachments) throws Exception {
        invisibleProps.put(ACTION, ACTION_SIGN);
        return getSignature(certificate, password, signText, visibleProps, invisibleProps,attachments);
    }

    /**
     * Get xmldsig signature message from parameters
     *
     * @param visibleProperties
     * @param invisibleProperties
     * @param attachments
     * @return StringBuffer
     * @throws java.security.NoSuchAlgorithmException
     *
     * @throws java.io.UnsupportedEncodingException
     *
     * @throws java.security.NoSuchProviderException
     *
     */
    private StringBuffer getSignature(AbstractCertificateHandler certificateHandler, char[] password, StringBuffer signText, Properties visibleProperties, Properties invisibleProperties, Attachment[] attachments) throws IOException, Exception {
        StringBuffer sb = new StringBuffer();
        sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n");
        sb.append("<openoces:signature xmlns:openoces=\""+NS_OPENOCES_SIGNATURE+"\" version=\"0.1\">");
        sb.append("<ds:Signature xmlns:ds=\""+NS_W3_XMLDSIG+"\">\n");

        StringBuffer objectText = getObject(visibleProperties, signText, invisibleProperties);
        StringBuffer[] attachmentObjects = getObjectsForAttachments(attachments);

        StringBuffer signedInfo = getSignedInfo(certificateHandler, objectText,attachmentObjects);

        sb.append(signedInfo.toString()).append("\n");
        sb.append(getSignatureValue(certificateHandler, password, signedInfo).toString());
        String keyInfo = getKeyInfo(certificateHandler, password);
        sb.append(keyInfo);
        sb.append(objectText.toString()).append("\n");

        if ( attachmentObjects != null ) {
            for (int i = 0; i < attachmentObjects.length; i++) {
                StringBuffer attachmentObject = attachmentObjects[i];
                if ( attachmentObject != null ) {
                    sb.append(attachmentObject.toString()).append("\n");
                }
            }
        }

        sb.append("</ds:Signature>");
        sb.append("</openoces:signature>");
        return sb;
    }

    private StringBuffer[] getObjectsForAttachments(Attachment[] attachments) {
        if ( attachments == null ) {
            return new StringBuffer[0];
        }
        
        StringBuffer sbs[] = new StringBuffer[attachments.length];

        for (int i = 0; i < attachments.length; i++) {
            Attachment attachment = attachments[i];
            if ( attachment.isHasSeen() ) {
                StringBuffer sb = new StringBuffer();
                sb.append("<ds:Object xmlns:ds=\""+NS_W3_XMLDSIG+"\"");
                sb.append(" xmlns:openoces=\""+NS_OPENOCES_SIGNATURE+"\"");
                sb.append(" Encoding=\"base64\"");
                sb.append(" Id=\"attachment-").append(i).append("\"");

                // FIXME: Instead of validating mime-types we should escape the mime-type appropriately.
                String mimeType = "text/plain";
                if ( MimeTypeValidator.validate(attachment.getMimeType())) {
                    mimeType = attachment.getMimeType();
                }

                sb.append(" MimeType=\"").append(mimeType).append("\"");
                sb.append(">");
                byte[] bs = attachment.getContents();
                if ( bs != null ) {
                    sb.append(wrapLines(new String(Base64.encode(bs)),BASE64_LINELENGTH).toString());
                }
                sb.append("</ds:Object>");
                sbs[i] = sb;
            } else {
                sbs[i] = null;
            }
        }
        return sbs;
    }

    /**
     * Get KEYINFO section of xmldsig
     *
     * @return String
     */
    private String getKeyInfo(AbstractCertificateHandler certificateHandler, char[] password) throws IOException, Exception {
        StringBuffer sb = new StringBuffer();
        sb.append("<ds:KeyInfo>\n");
        sb.append(getX509Data(certificateHandler, password));
        sb.append("</ds:KeyInfo>\n");
        return sb.toString();
    }

    /**
     * Get X509Data section of xmldsig
     *
     * @return String
     */
    private String getX509Data(AbstractCertificateHandler certificateHandler, char[] password) throws IOException, Exception {
        StringBuffer sb = new StringBuffer();
        byte[][] chain = certificateHandler.getCertificateChain(password);

        if (chain != null) {
            for (int i = 0; i < chain.length; i++) {
                sb.append("<ds:X509Data>\n");
                sb.append(getX509Certificate(chain[i]));
                sb.append("</ds:X509Data>\n");
            }
        }

        return sb.toString();
    }

    /**
     * Get X509Data section of xmldsig
     *
     * @param cert
     * @return String
     */
    private String getX509Certificate(byte[] cert) {
        StringBuffer sb = new StringBuffer();
        sb.append("<ds:X509Certificate>\n");
        sb.append(wrapLines(new String(Base64.encode(cert)), BASE64_LINELENGTH).toString());
        sb.append("\n</ds:X509Certificate>\n");
        return sb.toString();
    }

    /**
     * Get the Object section of the xmldsig message. Properties are encoded as NAME VALUE
     * pairs, where name and value are encoded ad BASE64, to avoid XML message syntaxt conflicts.
     * When decoded, signtext section sould be likewise decoded.
     * The invisible properties that have not been visible to the signer, whereas the visible
     * properties have been visible.
     *
     * @param visibleProperties
     * @param signText            the tbs text
     * @param invisibleProperties
     * @return StringBuffer
     */
    private StringBuffer getObject(Properties visibleProperties, StringBuffer signText, Properties invisibleProperties) throws UnsupportedEncodingException, InputDataError {
        StringBuffer sb = new StringBuffer();
        sb.append("<ds:Object xmlns:ds=\""+NS_W3_XMLDSIG+"\" xmlns:openoces=\""+NS_OPENOCES_SIGNATURE+"\" Id=\"ToBeSigned\">");
        sb.append("<ds:SignatureProperties>\n");
        sb.append(getPropertiesFragment(visibleProperties, signText, true).toString());
        sb.append(getPropertiesFragment(invisibleProperties, null, false).toString());
        sb.append("</ds:SignatureProperties>");
        sb.append("</ds:Object>");
        return sb;
    }

    /**
     * Get SignedInfo section of xmldsig
     *
     * @param toBeSigned
     * @param attachmentObjects
     * @return StringBuffer
     * @throws java.io.UnsupportedEncodingException
     *
     */
    private StringBuffer getSignedInfo(AbstractCertificateHandler certificateHandler, StringBuffer toBeSigned, StringBuffer[] attachmentObjects) throws Exception {
        StringBuffer sb = new StringBuffer();
        sb.append("<ds:SignedInfo xmlns:ds=\""+NS_W3_XMLDSIG+"\" xmlns:openoces=\""+NS_OPENOCES_SIGNATURE+"\">\n");
        sb.append("<ds:CanonicalizationMethod Algorithm=\"http://www.w3.org/TR/2001/REC-xml-c14n-20010315\"></ds:CanonicalizationMethod>\n");
        sb.append("<ds:SignatureMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#rsa-sha1\"></ds:SignatureMethod>\n");

        sb.append("<ds:Reference URI=\"#ToBeSigned\">\n");
        sb.append("<ds:DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"></ds:DigestMethod>\n");
        sb.append("<ds:DigestValue>");
        sb.append(getDigestValue(certificateHandler, toBeSigned));
        sb.append("</ds:DigestValue>\n");
        sb.append("</ds:Reference>\n");

        for (int i = 0; i < attachmentObjects.length; i++) {
            StringBuffer attachmentObject = attachmentObjects[i];
            if ( attachmentObject != null ) {
                sb.append("<ds:Reference Type=\"http://www.w3.org/2000/09/xmldsig#Object\" URI=\"#attachment-").append(i).append("\">\n");
                sb.append("<ds:DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"></ds:DigestMethod>\n");
                sb.append("<ds:DigestValue>");
                sb.append(getDigestValue(certificateHandler, attachmentObject));
                sb.append("</ds:DigestValue>\n");
                sb.append("</ds:Reference>\n");
            }
        }

        sb.append("</ds:SignedInfo>");
        return sb;
    }

    /**
     * Get Digest from signtext, coded as BASE64
     *
     * @param signText
     * @return String
     * @throws java.security.NoSuchProviderException
     *
     * @throws java.security.NoSuchAlgorithmException
     *
     * @throws UnsupportedEncodingException
     */
    private static String getDigestValue(AbstractCertificateHandler certificateHandler, StringBuffer signText) throws Exception {
        byte[] data = signText.toString().getBytes("UTF8");
        return new String(Base64.encode(certificateHandler.digest(data)));
    }

    /**
     * Get SignatureValue section of xmldsig
     *
     * @param signedInfo
     * @return String
F     * @throws java.security.InvalidKeyException
     *
     * @throws java.security.SignatureException
     *
     * @throws java.security.NoSuchAlgorithmException
     *
     * @throws UnsupportedEncodingException
     * @throws java.security.NoSuchProviderException
     *
     */
    private StringBuffer getSignatureValue(AbstractCertificateHandler certificateHandler, char[] password, StringBuffer signedInfo) throws IOException, Exception {
        StringBuffer sb = new StringBuffer();

        byte[] data = signedInfo.toString().getBytes("UTF8");
        FileLog.debug(HexDump.xdump(data));
        sb.append("<ds:SignatureValue>\n");
        sb.append(wrapLines(new String(Base64.encode(certificateHandler.sign(data, password))), BASE64_LINELENGTH).toString());
        sb.append("\n</ds:SignatureValue>\n");
        return sb;
    }

    private StringBuffer getPropertiesFragment(Properties properties, StringBuffer signText, boolean visible) throws UnsupportedEncodingException, InputDataError {
        StringBuffer s = new StringBuffer();
        if (signText != null) {
            s.append(getSignatureProperty("signtext", signText, true).toString()).append("\n");
        }

        if (properties != null) {
            Enumeration enumeration = properties.keys();
            while (enumeration.hasMoreElements()) {
                String name = enumeration.nextElement().toString();
                String value = properties.getProperty(name);
                s.append(getSignatureProperty(name, new StringBuffer(value), visible).toString()).append("\n");
            }
        }
        return s;
    }

    /**
     * Generate a <SignatureProperty> node, encode name and value in order to support special chars
     * s
     *
     * @param name
     * @param value
     * @param setToBeSignedId
     * @return the <SignatureProperty> node
     */

    private StringBuffer getSignatureProperty(String name, StringBuffer value, boolean setToBeSignedId) throws InputDataError, UnsupportedEncodingException {
        StringBuffer sb = new StringBuffer();

/*
            Better safe than sorry:
            An extra check in case we are not escaping all necessary characters correctly
        */
        if (!isWellformed(name)) {
            FileLog.warn("Name not wellformed: " + name);
            throw new InputDataError();
        }

        StringBuffer s;
        if (value == null || value.length() == 0) {
            s = new StringBuffer("");
        } else {
            s = new StringBuffer(new String(Base64.encode(value.toString().getBytes("UTF8"))));
        }

        sb.append("<ds:SignatureProperty>");
        sb.append("<openoces:Name>");
        sb.append(name);
        sb.append("</openoces:Name>");
        sb.append("<openoces:Value Encoding=\"base64\" VisibleToSigner=\"").append(setToBeSignedId ? "yes" : "no").append("\">");
        sb.append(s.toString());
        sb.append("</openoces:Value>");
        sb.append("</ds:SignatureProperty>");
        return sb;
    }


    private static boolean isWellformed(String name) {
        for (int i = 0; i < name.length(); i++) {
            int c = name.charAt(i);
            if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c == '_')) {
            } else {
                return false;
            }
        }
        return true;
    }

    /**
     * Do newline wrapping of a string
     *
     * @param s
     * @param len
     * @return the wrapped string
     */
    public static StringBuffer wrapLines(String s, int len) {

        /* invariant: points at the first character not yet copied */
        int sPos = 0;

        // sanity check
        if (s == null) {
            return null;
        }

        StringBuffer r = new StringBuffer();

        while (sPos < s.length()) {
            r.append(r.length() > 0 ? "\n" : "");
            if (s.length() - sPos <= len) {
                r.append(s.substring(sPos));
                sPos = s.length();
            } else {
                r.append(s.substring(sPos, sPos + len));
                sPos = sPos + len;
            }
        }
        return r;
    }
}